/// @file itc.cpp

#include <iostream>
#include <string>
#include <queue>
#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <unistd.h>
#include "itc.h"

using namespace std;
using namespace boost;
using namespace boost::posix_time;
using namespace itc;

// example #1 : new class which inherits from ITCThread 
class HelloThread : public ITCThread<HelloThread>
{
public:
	HelloThread(const string& name = "hello") :
		ITCThread<HelloThread>(name) {}
	
	void hello(const string& name)
	{
		doIt(&HelloThread::doHello, name);
	}

	void yahoo()
	{
		doIt(&HelloThread::doYahoo);
	}

private:
	void doHello(const string name)
	{
		cout << getName() << ": Hello~ " << name << endl;
	}
	
	void doYahoo()
	{
		cout << getName() << ": Yahoo~~~~~!" << endl;
	}
};

// example #2 : new class which inherits from ITCThread 
class PrintVector : public ITCThread<PrintVector>
{
public:
	PrintVector(const string& name = "prvec") :
		ITCThread<PrintVector>(name)
		{}
	
	void print(const vector<int>& vi)
	{
		doIt(&PrintVector::doPrint, vi);
	}

	void printTwoVectors(const vector<int>& vi, const vector<double>& vd)
	{
		doIt(&PrintVector::doPrintTwoVectors, vi, vd);
	}

private:
	void doPrint(const vector<int> vi)
	{
		copy(vi.begin(), vi.end(), ostream_iterator<int>(cout, " "));
		cout << endl;
	}

	void doPrintTwoVectors(const vector<int> vi, const vector<double> vd)
	{
		copy(vi.begin(), vi.end(), ostream_iterator<int>(cout, " "));
		cout << endl;
		copy(vd.begin(), vd.end(), ostream_iterator<double>(cout, " "));
		cout << endl;
	}
};

// example #3: legacy class + wrapping member function
struct Hello
{
	void sayHelloTo(const string name)
	{
		cout << "Hello, " << name << "!" << endl;
	}
};

struct HelloITCAdaptor : public ITCAdaptor<Hello>
{
	HelloITCAdaptor(const string& name) : ITCAdaptor<Hello>(name) {}

	void sayHelloTo(const string& name)
	{
		doIt(&Hello::sayHelloTo, name);
	}
};

// example #4: legacy class, no wrapping member function
struct ButtonHandler
{
	void onClick(int x, int y)
	{
		cout << "button clicked at (" << x << ", " << y << ")" << endl;
	}
};

typedef ITCAdaptor<ButtonHandler> ButtonHandlerITCAdaptor;

// example #5
void printHello()
{
	cout << "Hello, world" << endl;
}

void sayHelloTo(const string name)
{
	cout << "Hello, " << name << "!" << endl;
}

void sayWelcomeToAndGoodByeTo(const string name1, const string name2)
{
	cout << "Welcome, " << name1 << "!" << endl;
	cout << "Good-bye, " << name2 << "!" << endl;
}

// example #6
struct Greeter : public FunctionITCAdaptor<>
{
	Greeter(const string& name) : FunctionITCAdaptor<>(name) {}

	void printHello()
	{
		doIt(::printHello);
	}

	void sayHelloTo(const string& name)
	{
		doIt(::sayHelloTo, name);
	}

	void sayWelcomeToAndGoodByeTo(const string& name1, const string& name2)
	{
		doIt(::sayWelcomeToAndGoodByeTo, name1, name2);
	}
};

// example #7
class ControllerInternalProcessor : public ProcessAndExecute
{
private:
	virtual void process()
	{
		static int cnt = 0;
		if (cnt++ % 100 == 0)
		{
			cout << "processing..." << endl;
		}
		usleep(100000);		
	}
};

class Controller : public ITCThread<Controller, ControllerInternalProcessor>
{
public:
	Controller(const string& name) : 
		ITCThread<Controller, ControllerInternalProcessor>(name) 
	{}
	
	void setPosition(const string& name, double pos)
	{
		cout << getName() << ": queueing setPosition() request" << endl;
		doIt(&Controller::doSetPosition, name, pos);
	}

private:
	void doSetPosition(const string name, double pos)
	{
		cout << getName() << ": setting " << name
			 << "'s position to " << pos << endl;
	}
};

int
main()
{
#if 1
	string name;
	double pos;
	Controller ctrl("CTRL");
	
	do
	{
		cout << "enter name and position: ";
		cin >> name >> pos;
		cout << "setting " << name << "'s position to " << pos << endl;
		ctrl.setPosition(name, pos);
		usleep(100000);
	} while(cin);
	
	for (int i = 0; i < 100; ++i)
	{
		ctrl.setPosition(name, pos);
		usleep(10000);
	}
	ctrl.quit();
	ctrl.join();
#endif

#if 1
	// example #1
	HelloThread itc1("H1");
	HelloThread itc2("H2");
	
	for (int i = 0; i < 10; ++i)
	{
		this_thread::sleep(millisec(1000));
		itc1.hello("world!"), itc1.yahoo();
		this_thread::sleep(millisec(50));
		itc2.hello("body!"), itc2.yahoo();
	}
	this_thread::sleep(millisec(1000));
	itc1.quit(1);
	this_thread::sleep(millisec(50));
	itc2.quit(2);
	itc1.join(), itc2.join();

	// example #2
	string s = "prvec";
	PrintVector prvec(s);

	vector<int> vi;
	for (int i = 0; i < 10; ++i)
	{
		vi.push_back(i);
	}
	prvec.print(vi);
	this_thread::sleep(millisec(50));

	vector<double> vd;
	for (int i = 0; i < 10; ++i)
	{
		vd.push_back(i * 10.5);
	}
	prvec.printTwoVectors(vi, vd);

	prvec.quit();
	prvec.join();
	
	// example #3
	s = "hello";
	HelloITCAdaptor hello(s);
	for (int i = 0; i < 1000; ++i)
	{
		hello.sayHelloTo("Yoonsoo Kim");
	}
	hello.quit();
	hello.join();

	// example #4
	s = "btn_handler";
	ButtonHandlerITCAdaptor btnHdlr(s);
	btnHdlr.doIt(&ButtonHandler::onClick, 10, 20);
	btnHdlr.quit();
	btnHdlr.join();
	
	// example #5
	s = "fthr";
	FunctionITCAdaptor<> fthr(s);
	fthr.doIt(printHello);
	fthr.doIt(sayHelloTo, string("Yoonsoo Kim"));
	fthr.doIt(sayWelcomeToAndGoodByeTo, string("Yoonsoo Kim"), string("Yeongsuk Jeong"));
	fthr.quit();
	fthr.join();

	s = "GRTR";
	Greeter g(s);
	g.printHello();
	g.sayHelloTo("YYY");
	g.sayWelcomeToAndGoodByeTo("YYY", "KKK");
	g.quit();
	g.join();
#endif

	return 0;
}

